/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is The Browser Profile Migrator.
 *
 * The Initial Developer of the Original Code is Ben Goodger.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Ben Goodger <ben@bengoodger.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsBrowserProfileMigratorUtils.h"
#include "nsCRT.h"
#include "nsDirectoryServiceDefs.h"
#include "nsIObserverService.h"
#include "nsIPrefService.h"
#include "nsIRegistry.h"
#include "nsIServiceManager.h"
#include "nsISupportsArray.h"
#include "nsISupportsPrimitives.h"
#include "nsNetCID.h"
#include "nsNetUtil.h"
#include "nsFirefoxProfileMigrator.h"
#include "nsIProfileMigrator.h"
#include "nsIRDFDataSource.h"
#include "nsIRDFService.h"
#include "nsIExtensionManager.h"
#include "nsIPasswordManagerInternal.h"
#include "nsIPasswordManager.h"
#include "nsIPrefLocalizedString.h"
#include "nsIFormHistory.h"
#include "nsIMdbFactoryFactory.h"
#include "nsMorkCID.h"
//#include "nsIBrowserHistory.h"
//#include "nsIGlobalHistory.h"
#include "nsVoidArray.h"

#include <stdio.h>

#include "flockIIndexer.h"

#include "nsINIParser.h"

static void SwapBytes(PRUnichar* aDest, const PRUnichar* aSrc, PRUint32 aLen)
{
  for(PRUint32 i = 0; i < aLen; i++)
  {
    PRUnichar aChar = *aSrc++;
    *aDest++ = (0xff & (aChar >> 8)) | (aChar << 8);
  }
}


////////////////////////////////////////////////////////////////////////
//// Database I/O

class MigrationErrorHook : public nsIMdbErrorHook
{
public:
  NS_DECL_ISUPPORTS

  // nsIMdbErrorHook
  NS_IMETHOD OnErrorString(nsIMdbEnv* ev, const char* inAscii);
  NS_IMETHOD OnErrorYarn(nsIMdbEnv* ev, const mdbYarn* inYarn);
  NS_IMETHOD OnWarningString(nsIMdbEnv* ev, const char* inAscii);
  NS_IMETHOD OnWarningYarn(nsIMdbEnv* ev, const mdbYarn* inYarn);
  NS_IMETHOD OnAbortHintString(nsIMdbEnv* ev, const char* inAscii);
  NS_IMETHOD OnAbortHintYarn(nsIMdbEnv* ev, const mdbYarn* inYarn);
};

// nsIMdbErrorHook has no IID!
NS_IMPL_ISUPPORTS0(MigrationErrorHook)

NS_IMETHODIMP
MigrationErrorHook::OnErrorString(nsIMdbEnv *ev, const char *inAscii)
{
  printf("mork error: %s\n", inAscii);
  return NS_OK;
}

NS_IMETHODIMP
MigrationErrorHook::OnErrorYarn(nsIMdbEnv *ev, const mdbYarn* inYarn)
{
  printf("mork error yarn: %p\n", (void*)inYarn);
  return NS_OK;
}

NS_IMETHODIMP
MigrationErrorHook::OnWarningString(nsIMdbEnv *ev, const char *inAscii)
{
  printf("mork warning: %s\n", inAscii);
  return NS_OK;
}

NS_IMETHODIMP
MigrationErrorHook::OnWarningYarn(nsIMdbEnv *ev, const mdbYarn *inYarn)
{
  printf("mork warning yarn: %p\n", (void*)inYarn);
  return NS_OK;
}

NS_IMETHODIMP
MigrationErrorHook::OnAbortHintString(nsIMdbEnv *ev, const char *inAscii)
{
  printf("mork abort: %s\n", inAscii);
  return NS_OK;
}

NS_IMETHODIMP
MigrationErrorHook::OnAbortHintYarn(nsIMdbEnv *ev, const mdbYarn *inYarn)
{
  printf("mork abort yarn: %p\n", (void*)inYarn);
  return NS_OK;
}


///////////////////////////////////////////////////////////////////////////////
// nsFirefoxProfileMigrator

#define FILE_NAME_BOOKMARKS       NS_LITERAL_STRING("bookmarks.html")
#define FILE_NAME_COOKIES         NS_LITERAL_STRING("cookies.txt")
#define FILE_NAME_SITEPERM_OLD    NS_LITERAL_STRING("cookperm.txt")
#define FILE_NAME_SITEPERM_NEW    NS_LITERAL_STRING("hostperm.1")
#define FILE_NAME_CERT8DB         NS_LITERAL_STRING("cert8.db")
#define FILE_NAME_KEY3DB          NS_LITERAL_STRING("key3.db")
#define FILE_NAME_KEY3DB2         NS_LITERAL_STRING("key3.db2")
#define FILE_NAME_SECMODDB        NS_LITERAL_STRING("secmod.db")
#define FILE_NAME_HISTORY         NS_LITERAL_STRING("history.dat")
#define FILE_NAME_FORMHISTORY     NS_LITERAL_STRING("formhistory.dat")
#define FILE_NAME_LOCALSTORE      NS_LITERAL_STRING("localstore.rdf")
#define FILE_NAME_MIMETYPES       NS_LITERAL_STRING("mimeTypes.rdf")
#define FILE_NAME_DOWNLOADS       NS_LITERAL_STRING("downloads.rdf")
#define FILE_NAME_PREFS           NS_LITERAL_STRING("prefs.js")
#define FILE_NAME_USER_PREFS      NS_LITERAL_STRING("user.js")
#define FILE_NAME_SEARCH          NS_LITERAL_STRING("search.rdf")
#define FILE_NAME_USERCHROME      NS_LITERAL_STRING("userChrome.css")
#define FILE_NAME_USERCONTENT     NS_LITERAL_STRING("userContent.css")
#define DIR_NAME_CHROME           NS_LITERAL_STRING("chrome")
#define DIR_NAME_EXTENSIONS       NS_LITERAL_STRING("extensions")

#define INSTALL_MANIFEST          NS_LITERAL_CSTRING("urn:mozilla:install-manifest")
#define EMNAME_RESOURCE           NS_LITERAL_CSTRING("http://www.mozilla.org/2004/em-rdf#name")

#define NETSCAPE8_FILENAME        "netscape.exe"

NS_IMPL_ISUPPORTS1(nsFirefoxProfileMigrator, nsIBrowserProfileMigrator)

nsFirefoxProfileMigrator::nsFirefoxProfileMigrator() :
  mEnv(nsnull),
  mStore(nsnull),
  mTable(nsnull),
  mReverseByteOrder(PR_FALSE)
{
  mObserverService = do_GetService("@mozilla.org/observer-service;1");
}

nsFirefoxProfileMigrator::~nsFirefoxProfileMigrator()
{
}

nsresult
nsFirefoxProfileMigrator::CopyFile(const nsAString& aSourceFileName, const nsAString& aTargetFileName)
{
  nsCOMPtr<nsIFile> sourceFile;
  mSourceProfile->Clone(getter_AddRefs(sourceFile));

  sourceFile->Append(aSourceFileName);
  PRBool exists = PR_FALSE;
  sourceFile->Exists(&exists);
  if (!exists)
    return NS_OK;

  nsCOMPtr<nsIFile> targetFile;
  mTargetProfile->Clone(getter_AddRefs(targetFile));
  
  targetFile->Append(aTargetFileName);
  targetFile->Exists(&exists);
  if (exists)
    targetFile->Remove(PR_FALSE);

  return sourceFile->CopyTo(mTargetProfile, aTargetFileName);
}

PRBool
nsFirefoxProfileMigrator::SourceFileExists(const nsAString& aSourceFileName)
{
  nsCOMPtr<nsIFile> sourceFile;
  mSourceProfile->Clone(getter_AddRefs(sourceFile));
  sourceFile->Append(aSourceFileName);
  PRBool exists = PR_FALSE;
  sourceFile->Exists(&exists);
  return exists;
 }

///////////////////////////////////////////////////////////////////////////////
// nsIBrowserProfileMigrator

NS_IMETHODIMP
nsFirefoxProfileMigrator::Migrate(PRUint16 aItems, nsIProfileStartup* aStartup, const PRUnichar* aProfile)
{
  printf("****** Starting nsFirefoxProfileMigrator::Migrate()\n");
  nsresult rv = NS_OK;
  PRBool aReplace = aStartup ? PR_TRUE : PR_FALSE;

  if (!mTargetProfile) {
    printf("****** nsFirefoxProfileMigrator::Migrate() - Getting Target Profile\n");
    GetProfilePath(aStartup, mTargetProfile);
    if (!mTargetProfile) return NS_ERROR_FAILURE;
  }

  if (!mSourceProfile) {
    printf("****** nsFirefoxProfileMigrator::Migrate() - Getting Source Profile\n");
    GetSourceProfile(aProfile);
  }

  NOTIFY_OBSERVERS(MIGRATION_STARTED, nsnull);

  COPY_DATA(CopyPreferences,  aReplace, nsIBrowserProfileMigrator::SETTINGS);     // Settings are now migrated properly (only needed on first run)
  COPY_DATA(CopyCookies,      aReplace, nsIBrowserProfileMigrator::COOKIES);      // Will merge on subsequent imports.
  COPY_DATA(CopyHistory,      aReplace, nsIBrowserProfileMigrator::HISTORY);
  COPY_DATA(CopyFormHistory,  aReplace, nsIBrowserProfileMigrator::FORMDATA);
  COPY_DATA(CopyOtherData,    aReplace, nsIBrowserProfileMigrator::OTHERDATA);

  if (aReplace &&
      (aItems & nsIBrowserProfileMigrator::SETTINGS || 
       aItems & nsIBrowserProfileMigrator::COOKIES || 
       aItems & nsIBrowserProfileMigrator::PASSWORDS ||
       !aItems)) {
    // Permissions (Images, Cookies, Popups)
    printf("****** nsFirefoxProfileMigrator::Migrate() - Migrating Permissions\n");
    rv |= CopyFile(FILE_NAME_SITEPERM_NEW, FILE_NAME_SITEPERM_NEW);
    rv |= CopyFile(FILE_NAME_SITEPERM_OLD, FILE_NAME_SITEPERM_OLD);
  }

  // Enable access to ProfD for dependent services (ie favorites)
  if (aStartup) {
    printf("****** nsFirefoxProfileMigrator::Migrate() - Detected first run so doing startup\n");
    rv = aStartup->DoStartup();
    NS_ENSURE_SUCCESS(rv, rv);
  }
  
  COPY_DATA(CopyPasswords,    aReplace, nsIBrowserProfileMigrator::PASSWORDS);    // Will merge on subsequent imports.
  COPY_DATA(CopyBookmarks,    PR_FALSE, nsIBrowserProfileMigrator::BOOKMARKS);    // Will merge on subsequent imports.

  printf("****** nsFirefoxProfileMigrator::Migrate() - Rebuilding Index\n");
  // tell indexer to rebuild index for new history
  nsCOMPtr<flockIIndexer> indexer(do_GetService("@flock.com/indexer;1", &rv));
  if (NS_FAILED(rv)) return rv;
  rv = indexer->RebuildIndex();
  if (NS_FAILED(rv)) return rv;  
  
  printf("****** nsFirefoxProfileMigrator::Migrate() - Migrating Yahoo! Toolbar\n");
  // Added in an extra import for the yahoo toolbar.
  MigrateYahooToolbar();

  printf("****** nsFirefoxProfileMigrator::Migrate() - Finishing C++ part of Migration\n");
  NOTIFY_OBSERVERS(MIGRATION_ENDED, nsnull);

  return rv;
}

NS_IMETHODIMP
nsFirefoxProfileMigrator::GetMigrateData(const PRUnichar* aProfile, 
                                         PRBool aReplace, 
                                         PRUint16* aResult)
{
  *aResult = 0;
  if (!mSourceProfile) {
    GetSourceProfile(aProfile);
    if (!mSourceProfile)
      return NS_ERROR_FILE_NOT_FOUND;
  }

  /*
    Struct values and semantics:

    PRUnichar* fileName :   Name of file on disk, to check whether source exists.
    PRUint32 sourceFlag :   The bit to set in aResult if element should be migrated.
    PRBool replaceOnly  :   If TRUE, omit element unless aReplace is also TRUE.
  */
  MigrationData data[] = { 
                           { ToNewUnicode(FILE_NAME_PREFS),
                             nsIBrowserProfileMigrator::SETTINGS,
                             PR_TRUE },
                           { ToNewUnicode(FILE_NAME_USER_PREFS),
                             nsIBrowserProfileMigrator::SETTINGS,
                             PR_TRUE },
                           { ToNewUnicode(FILE_NAME_COOKIES),
                             nsIBrowserProfileMigrator::COOKIES,
                             PR_FALSE },
                           { ToNewUnicode(FILE_NAME_HISTORY),
                             nsIBrowserProfileMigrator::HISTORY,
                             PR_TRUE },   // TODO: Implement merge, and set to PR_FALSE.
                           { ToNewUnicode(FILE_NAME_BOOKMARKS),
                             nsIBrowserProfileMigrator::BOOKMARKS,
                             PR_FALSE },
                           { ToNewUnicode(FILE_NAME_DOWNLOADS),
                             nsIBrowserProfileMigrator::OTHERDATA,
                             PR_TRUE },
                           { ToNewUnicode(FILE_NAME_MIMETYPES),
                             nsIBrowserProfileMigrator::OTHERDATA,
                             PR_TRUE },
                           { ToNewUnicode(FILE_NAME_USERCHROME),
                             nsIBrowserProfileMigrator::OTHERDATA,
                             PR_TRUE },
                           { ToNewUnicode(FILE_NAME_USERCONTENT),
                             nsIBrowserProfileMigrator::OTHERDATA,
                             PR_TRUE },
                           { ToNewUnicode(FILE_NAME_FORMHISTORY),
                             nsIBrowserProfileMigrator::FORMDATA,
                             PR_FALSE },
                           { ToNewUnicode(FILE_NAME_DOWNLOADS),
                             nsIBrowserProfileMigrator::OTHERDATA,
                             PR_TRUE } };

  // Frees file name strings allocated above.
  GetMigrateDataFromArray(data, sizeof(data)/sizeof(MigrationData), 
                          aReplace, mSourceProfile, aResult);

  // Now locate the passwords.
  nsXPIDLCString signonsFileName;
  GetSourceSignonFileName(aReplace, getter_Copies(signonsFileName));
  if (!signonsFileName.IsEmpty()) {
    nsAutoString fileName;
    fileName.AssignWithConversion(signonsFileName);
    if (SourceFileExists(fileName)) {
      *aResult |= nsIBrowserProfileMigrator::PASSWORDS;
      // Save file name so we don't have to get it again later.
      mSourceSignonFileName = fileName;
    }
  }    

  return NS_OK;
}

nsresult nsFirefoxProfileMigrator::GetSourceSignonFileName(PRBool aReplace,
                                                           char** aFileName)
{
  return GetSignonFileName(aReplace, aFileName);
}

NS_IMETHODIMP
nsFirefoxProfileMigrator::GetSourceExists(PRBool* aResult)
{
  nsCOMPtr<nsISupportsArray> profiles;
  GetSourceProfiles(getter_AddRefs(profiles));

  if (profiles) { 
    PRUint32 count;
    profiles->Count(&count);
    *aResult = count > 0;
  }
  else
    *aResult = PR_FALSE;

  return NS_OK;
}

NS_IMETHODIMP
nsFirefoxProfileMigrator::GetSourceHasMultipleProfiles(PRBool* aResult)
{
  nsCOMPtr<nsISupportsArray> profiles;
  GetSourceProfiles(getter_AddRefs(profiles));

  if (profiles) {
    PRUint32 count;
    profiles->Count(&count);
    *aResult = count > 1;
  }
  else
    *aResult = PR_FALSE;

  return NS_OK;
}

NS_IMETHODIMP
nsFirefoxProfileMigrator::GetSourceProfiles(nsISupportsArray** aResult)
{
  if (!mProfileNames && !mProfileLocations) {
    nsresult rv = NS_NewISupportsArray(getter_AddRefs(mProfileNames));
    if (NS_FAILED(rv)) return rv;

    rv = NS_NewISupportsArray(getter_AddRefs(mProfileLocations));
    if (NS_FAILED(rv)) return rv;

    // Fills mProfileNames and mProfileLocations
    FillProfileDataFromRegistry();
  }
  
  NS_IF_ADDREF(*aResult = mProfileNames);
  return NS_OK;
}

NS_IMETHODIMP
nsFirefoxProfileMigrator::GetSourceHomePageURL(nsACString& aResult)
{
  nsresult rv;
  nsCOMPtr<nsIPrefService> psvc(do_GetService(NS_PREFSERVICE_CONTRACTID));
  psvc->ResetPrefs();

  nsCOMPtr<nsIFile> seamonkeyPrefsFile;
  mSourceProfile->Clone(getter_AddRefs(seamonkeyPrefsFile));
  seamonkeyPrefsFile->Append(FILE_NAME_PREFS);
  psvc->ReadUserPrefs(seamonkeyPrefsFile);

  nsXPIDLCString homePage;
  nsCOMPtr<nsIPrefBranch> branch(do_QueryInterface(psvc));
  rv = branch->GetCharPref("browser.startup.homepage", getter_Copies(homePage));
  if(!NS_SUCCEEDED(rv)) return rv;
  
  if (homePage.EqualsLiteral("resource:/browserconfig.properties")) {
    nsAutoString nativePath, fileName;

    rv = GetAppFileName(fileName);
    NS_ENSURE_SUCCESS(rv, rv);

#ifdef XP_WIN32
    // The preference for browser.startup.homepage may be a resource
    // ie: resource:///browserconfig.properties
    // We need to load up browserconfig.properties and read in the line with
    // "browser.startup.homepage=..."
    // First we need the application install location (since that is where
    // browserconfig.properties is located)
    nsAutoString appPathKey;
    appPathKey.AssignLiteral("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\");
    appPathKey.Append(fileName);

    nsCOMPtr<nsIWindowsRegKey> regKey = do_CreateInstance("@mozilla.org/windows-registry-key;1");
    rv = regKey->Open(nsIWindowsRegKey::ROOT_KEY_LOCAL_MACHINE, appPathKey,
                      nsIWindowsRegKey::ACCESS_READ);
    NS_ENSURE_SUCCESS(rv, rv);

    rv = regKey->ReadStringValue(NS_LITERAL_STRING("Path"), nativePath);
    NS_ENSURE_SUCCESS(rv, rv);

    nativePath.AppendLiteral("\\browserconfig.properties");
    nativePath.Insert(NS_LITERAL_STRING("file://"), 0);

#elif defined(XP_MACOSX)
    nsAutoString appName(fileName);
    appName.AppendLiteral(".app");

    // To find the application directory associated with this profile, we
    // look in two places:
    // /Applications/<appname>/Contents/MacOS
    // ~/Desktop/<appname>/Contents/MacOS
    nsCOMPtr<nsIProperties> fileLocator(do_GetService("@mozilla.org/file/directory_service;1", &rv));
    NS_ENSURE_SUCCESS(rv, rv);

    nsCOMPtr<nsILocalFile> appDir;
    rv = fileLocator->Get(NS_OSX_LOCAL_APPLICATIONS_DIR,
                          NS_GET_IID(nsILocalFile),
                          getter_AddRefs(appDir));
    NS_ENSURE_SUCCESS(rv, rv);

    appDir->Append(appName);

    PRBool exists = PR_FALSE;
    appDir->Exists(&exists);
    if (!exists) {
      rv = fileLocator->Get(NS_OSX_USER_DESKTOP_DIR,
                            NS_GET_IID(nsILocalFile),
                            getter_AddRefs(appDir));
      NS_ENSURE_SUCCESS(rv, rv);

      appDir->Append(appName);

      appDir->Exists(&exists);
      if (!exists) {
        nsAutoString volumePath;

        volumePath.AppendLiteral("/Volumes/");
        volumePath.Append(fileName);
        volumePath.AppendLiteral("/");
        volumePath.Append(appName);

        appDir = do_CreateInstance(NS_LOCAL_FILE_CONTRACTID, &rv);
        NS_ENSURE_SUCCESS(rv, rv);

        appDir->InitWithPath(volumePath);
      }
    }

    appDir->GetPath(nativePath);
    nativePath.AppendLiteral("/Contents/MacOS/browserconfig.properties");
    nativePath.Insert(NS_LITERAL_STRING("file://"), 0);
#else
    aResult.Assign(homePage);
    return NS_OK;
#endif
    // Use the StringBundleService to parse the browserconfig.properties file
    nsCOMPtr<nsIStringBundleService> bundleService = do_GetService(NS_STRINGBUNDLE_CONTRACTID, &rv);
    if (NS_FAILED(rv)) return rv;
    nsCOMPtr<nsIStringBundle> entities;
    rv = bundleService->CreateBundle(NS_ConvertUTF16toUTF8(nativePath).get(), getter_AddRefs(entities));
    if (NS_FAILED(rv)) return rv;
    nsXPIDLString value;
    rv = entities->GetStringFromName(NS_LITERAL_STRING("browser.startup.homepage").get(), getter_Copies(value));
    if (NS_FAILED(rv)) return rv;
    CopyUTF16toUTF8(nsAutoString(value), aResult);
  } else {
    aResult.Assign(homePage);
  }
  
  return NS_OK;
}
///////////////////////////////////////////////////////////////////////////////
// nsFirefoxProfileMigrator

nsresult
nsFirefoxProfileMigrator::GetSourceProfile(const PRUnichar* aProfile)
{
  PRUint32 count;
  mProfileNames->Count(&count);
  for (PRUint32 i = 0; i < count; ++i) {
    nsCOMPtr<nsISupportsString> str(do_QueryElementAt(mProfileNames, i));
    nsXPIDLString profileName;
    str->GetData(profileName);
    if (profileName.Equals(aProfile)) {
      mSourceProfile = do_QueryElementAt(mProfileLocations, i);
      break;
    }
  }

  return NS_OK;
}

nsresult nsFirefoxProfileMigrator::AppendNativeProfilePath(nsILocalFile *aFile) {
#ifdef XP_WIN
  aFile->Append(NS_LITERAL_STRING("Mozilla"));
  aFile->Append(NS_LITERAL_STRING("Firefox"));
#elif defined(XP_MACOSX)
  aFile->Append(NS_LITERAL_STRING("Application Support"));
  aFile->Append(NS_LITERAL_STRING("Firefox"));
#elif defined(XP_UNIX)
  aFile->Append(NS_LITERAL_STRING(".mozilla"));
  aFile->Append(NS_LITERAL_STRING("firefox"));
#endif
  return NS_OK;
}

nsresult
nsFirefoxProfileMigrator::FillProfileDataFromRegistry()
{
  nsCOMPtr<nsIProperties> fileLocator(do_GetService("@mozilla.org/file/directory_service;1"));
  nsCOMPtr<nsILocalFile> firefoxRegistry;
  nsCOMPtr<nsILocalFile> profileRoot;
  char *APPDATADIR;
#ifdef XP_WIN
  APPDATADIR = NS_WIN_APPDATA_DIR;
#elif defined(XP_MACOSX)
  APPDATADIR = NS_MAC_USER_LIB_DIR;
#elif defined(XP_UNIX)
  APPDATADIR = NS_UNIX_HOME_DIR;
#endif
  fileLocator->Get(APPDATADIR, NS_GET_IID(nsILocalFile), getter_AddRefs(firefoxRegistry));
  fileLocator->Get(APPDATADIR, NS_GET_IID(nsILocalFile), getter_AddRefs(profileRoot));

  AppendNativeProfilePath(profileRoot);
  AppendNativeProfilePath(firefoxRegistry);

  nsCAutoString test;
   
  firefoxRegistry->AppendNative(NS_LITERAL_CSTRING("profiles.ini"));
  //firefoxRegistry->GetNativePath(test);
  //printf("YOO %s\n", test.get());
  nsINIParser parser;
  nsresult rv = parser.Init(firefoxRegistry);
  if (NS_FAILED(rv)) return rv;

  unsigned int c = 0;
  nsCAutoString buffer;
  for (c = 0; PR_TRUE; ++c) {
      nsCAutoString profileID("Profile");
      profileID.AppendInt(c);

      rv = parser.GetString(profileID.get(), "IsRelative", buffer);
      if (NS_FAILED(rv)) break;

      PRBool isRelative = buffer.EqualsLiteral("1");

      nsCAutoString filePath;
      nsCAutoString profileName;

      rv = parser.GetString(profileID.get(), "Path", filePath);
      if (NS_FAILED(rv)) {
          NS_ERROR("Malformed profiles.ini: Path= not found");
          continue;
      }

      rv = parser.GetString(profileID.get(), "Name", profileName);
      if (NS_FAILED(rv)) {
          NS_ERROR("Malformed profiles.ini: Name= not found");
          continue;
      }

      nsCOMPtr<nsILocalFile> firefoxProfileDir;
      rv = NS_NewNativeLocalFile(EmptyCString(), PR_TRUE,
                                  getter_AddRefs(firefoxProfileDir));
      NS_ENSURE_SUCCESS(rv, rv);
      if (isRelative) {
          rv = firefoxProfileDir->SetRelativeDescriptor(profileRoot, filePath);
      } else {
          rv = firefoxProfileDir->SetPersistentDescriptor(filePath);
      }
      if (NS_FAILED(rv)) continue;
      PRBool exists;
      firefoxProfileDir->Exists(&exists);
      nsAutoString test;
          rv = firefoxProfileDir->GetPath(test);

      if (exists) {
        nsCOMPtr<nsISupportsString> profileNameString(do_CreateInstance("@mozilla.org/supports-string;1"));
        profileNameString->SetData(NS_ConvertUTF8toUTF16(profileName));
        mProfileNames->AppendElement(profileNameString);

        mProfileLocations->AppendElement(firefoxProfileDir);
      }
/*
      nsCOMPtr<nsILocalFile> rootDir;
      rv = NS_NewNativeLocalFile(EmptyCString(), PR_TRUE,
                                  getter_AddRefs(rootDir));
      NS_ENSURE_SUCCESS(rv, rv);

      if (isRelative) {
          rv = rootDir->SetRelativeDescriptor(mAppData, filePath);
      } else {
          rv = rootDir->SetPersistentDescriptor(filePath);
      }
      if (NS_FAILED(rv)) continue;

      nsCOMPtr<nsILocalFile> localDir;
      if (isRelative) {
          rv = NS_NewNativeLocalFile(EmptyCString(), PR_TRUE,
                                      getter_AddRefs(localDir));
          NS_ENSURE_SUCCESS(rv, rv);

          rv = localDir->SetRelativeDescriptor(mTempData, filePath);
      } else {
          localDir = rootDir;
      }
      nsCOMPtr<nsILocalFile> dir;

      currentProfile = new nsToolkitProfile(buffer,
                                            rootDir, localDir,
                                            currentProfile);
      NS_ENSURE_TRUE(currentProfile, NS_ERROR_OUT_OF_MEMORY);

      rv = parser.GetString(profileID.get(), "Default", buffer);
      if (NS_SUCCEEDED(rv) && buffer.EqualsLiteral("1"))
          mChosen = currentProfile;
*/
  }
  return NS_OK;
/*
  // Find the Firefox Registry
  nsCOMPtr<nsIProperties> fileLocator(do_GetService("@mozilla.org/file/directory_service;1"));
  nsCOMPtr<nsILocalFile> firefoxRegistry;
#ifdef XP_WIN
  fileLocator->Get(NS_WIN_APPDATA_DIR, NS_GET_IID(nsILocalFile), getter_AddRefs(firefoxRegistry));

  firefoxRegistry->Append(NS_LITERAL_STRING("Firefox"));
  firefoxRegistry->Append(NS_LITERAL_STRING("registry.dat"));
#elif defined(XP_MACOSX)
  fileLocator->Get(NS_MAC_USER_LIB_DIR, NS_GET_IID(nsILocalFile), getter_AddRefs(firefoxRegistry));
  
  firefoxRegistry->Append(NS_LITERAL_STRING("Firefox"));
  firefoxRegistry->Append(NS_LITERAL_STRING("Application Registry"));
#elif defined(XP_UNIX)
  fileLocator->Get(NS_UNIX_HOME_DIR, NS_GET_IID(nsILocalFile), getter_AddRefs(firefoxRegistry));
  
  firefoxRegistry->Append(NS_LITERAL_STRING(".firefox"));
  firefoxRegistry->Append(NS_LITERAL_STRING("appreg"));
#elif defined(XP_OS2)
  fileLocator->Get(NS_OS2_HOME_DIR, NS_GET_IID(nsILocalFile), getter_AddRefs(firefoxRegistry));
  
  firefoxRegistry->Append(NS_LITERAL_STRING("Firefox"));
  firefoxRegistry->Append(NS_LITERAL_STRING("registry.dat"));
#endif


  return GetProfileDataFromRegistry(firefoxRegistry, mProfileNames, mProfileLocations);
*/
}

#define F(a) nsFirefoxProfileMigrator::a

#define MAKEPREFTRANSFORM(pref, newpref, getmethod, setmethod) \
  { pref, newpref, F(Get##getmethod), F(Set##setmethod), PR_FALSE, { -1 } }

#define MAKESAMETYPEPREFTRANSFORM(pref, method) \
  { pref, 0, F(Get##method), F(Set##method), PR_FALSE, { -1 } }


static 
nsFirefoxProfileMigrator::PrefTransform gTransforms[] = {
  MAKESAMETYPEPREFTRANSFORM("signon.SignonFileName",                    String),
  MAKESAMETYPEPREFTRANSFORM("signon.SignonFileName2",                   String),
  MAKESAMETYPEPREFTRANSFORM("browser.history_expire_days",              Int),
  MAKESAMETYPEPREFTRANSFORM("browser.tabs.autoHide",                    Bool),
  MAKESAMETYPEPREFTRANSFORM("browser.tabs.loadInBackground",            Bool),
  MAKESAMETYPEPREFTRANSFORM("browser.enable_automatic_image_resizing",  Bool),
  MAKESAMETYPEPREFTRANSFORM("network.cookie.warnAboutCookies",          Bool),
  MAKESAMETYPEPREFTRANSFORM("network.cookie.lifetime.enabled",          Bool),
  MAKESAMETYPEPREFTRANSFORM("network.cookie.lifetime.behavior",         Int),
  MAKESAMETYPEPREFTRANSFORM("dom.disable_open_during_load",             Bool),
  MAKESAMETYPEPREFTRANSFORM("signon.rememberSignons",                   Bool),
  MAKESAMETYPEPREFTRANSFORM("security.enable_ssl2",                     Bool),
  MAKESAMETYPEPREFTRANSFORM("security.enable_ssl3",                     Bool),
  MAKESAMETYPEPREFTRANSFORM("security.enable_tls",                      Bool),
  MAKESAMETYPEPREFTRANSFORM("security.warn_entering_secure",            Bool),
  MAKESAMETYPEPREFTRANSFORM("security.warn_entering_weak",              Bool),
  MAKESAMETYPEPREFTRANSFORM("security.warn_leaving_secure",             Bool),
  MAKESAMETYPEPREFTRANSFORM("security.warn_submit_insecure",            Bool),
  MAKESAMETYPEPREFTRANSFORM("security.warn_viewing_mixed",              Bool),
  MAKESAMETYPEPREFTRANSFORM("security.default_personal_cert",           String),
  MAKESAMETYPEPREFTRANSFORM("security.OSCP.enabled",                    Int),
  MAKESAMETYPEPREFTRANSFORM("security.OSCP.signingCA",                  String),
  MAKESAMETYPEPREFTRANSFORM("security.OSCP.URL",                        String),
  MAKESAMETYPEPREFTRANSFORM("security.enable_java",                     Bool),
  MAKESAMETYPEPREFTRANSFORM("javascript.enabled",                       Bool),
  MAKESAMETYPEPREFTRANSFORM("dom.disable_window_move_resize",           Bool),
  MAKESAMETYPEPREFTRANSFORM("dom.disable_window_flip",                  Bool),
  MAKESAMETYPEPREFTRANSFORM("dom.disable_window_open_feature.status",   Bool),
  MAKESAMETYPEPREFTRANSFORM("dom.disable_window_status_change",         Bool),
  MAKESAMETYPEPREFTRANSFORM("dom.disable_image_src_set",                Bool),
  MAKESAMETYPEPREFTRANSFORM("accessibility.typeaheadfind.autostart",    Bool),
  MAKESAMETYPEPREFTRANSFORM("accessibility.typeaheadfind.linksonly",    Bool),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.type",                       Int),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.http",                       String),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.http_port",                  Int),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.ftp",                        String),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.ftp_port",                   Int),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.ssl",                        String),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.ssl_port",                   Int),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.socks",                      String),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.socks_port",                 Int),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.gopher",                     String),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.gopher_port",                Int),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.no_proxies_on",              String),
  MAKESAMETYPEPREFTRANSFORM("network.proxy.autoconfig_url",             String),
  MAKESAMETYPEPREFTRANSFORM("browser.display.foreground_color",         String),
  MAKESAMETYPEPREFTRANSFORM("browser.display.background_color",         String),
  MAKESAMETYPEPREFTRANSFORM("browser.anchor_color",                     String),
  MAKESAMETYPEPREFTRANSFORM("browser.visited_color",                    String),
  MAKESAMETYPEPREFTRANSFORM("browser.underline_anchors",                Bool),
  MAKESAMETYPEPREFTRANSFORM("browser.display.use_system_colors",        Bool),
  MAKESAMETYPEPREFTRANSFORM("browser.display.use_document_colors",      Bool),
  MAKESAMETYPEPREFTRANSFORM("browser.display.use_document_fonts",       Bool),
  MAKESAMETYPEPREFTRANSFORM("intl.charset.default",                     String),
  MAKESAMETYPEPREFTRANSFORM("intl.accept_languages",                    String),
  MAKESAMETYPEPREFTRANSFORM("intl.accept_charsets",                     String),

  MAKEPREFTRANSFORM("network.image.imageBehavior",      0, Int, Image),
  MAKEPREFTRANSFORM("network.cookie.cookieBehavior",    0, Int, Cookie),
  MAKEPREFTRANSFORM("browser.downloadmanager.behavior", 0, Int, DownloadManager),

  MAKEPREFTRANSFORM("wallet.captureForms", "formfill.enabled", Bool, Bool)
};

nsresult 
nsFirefoxProfileMigrator::SetImage(void* aTransform, nsIPrefBranch* aBranch)
{
  PrefTransform* xform = (PrefTransform*)aTransform;
  nsresult rv = NS_OK;

  if (xform->prefHasValue)
    rv = aBranch->SetIntPref("network.image.imageBehavior", xform->intValue == 1 ? 0 : xform->intValue);

  return rv;
}

nsresult 
nsFirefoxProfileMigrator::SetCookie(void* aTransform, nsIPrefBranch* aBranch)
{
  PrefTransform* xform = (PrefTransform*)aTransform;
  nsresult rv = NS_OK;

  if (xform->prefHasValue)
    rv = aBranch->SetIntPref("network.cookie.cookieBehavior", xform->intValue == 3 ? 0 : xform->intValue);

  return rv;
}

nsresult 
nsFirefoxProfileMigrator::SetDownloadManager(void* aTransform, nsIPrefBranch* aBranch)
{
  PrefTransform* xform = (PrefTransform*)aTransform;
  nsresult rv = NS_OK;
  
  if (xform->prefHasValue) {
    // Seamonkey's download manager uses a single pref to control behavior:
    // 0 - show download manager window
    // 1 - show individual progress dialogs
    // 2 - show nothing
    //
    // Firebird has only a download manager window, but it can behave like a progress dialog, thus:
    // 0 || 1  -> show downloads window when a download starts
    // 2       -> don't show anything when a download starts
    // 1       -> close the downloads window as if it were a progress window when downloads complete.
    //
    rv |= aBranch->SetBoolPref("browser.download.manager.showWhenStarting", xform->intValue != 2);
    rv |= aBranch->SetBoolPref("browser.download.manager.closeWhenDone", xform->intValue == 1);
  }
  return NS_OK;
}

nsresult
nsFirefoxProfileMigrator::TransformPreferences(const nsAString& aSourcePrefFileName,
                                                 const nsAString& aTargetPrefFileName)
{
  PrefTransform* transform;
  PrefTransform* end = gTransforms + sizeof(gTransforms)/sizeof(PrefTransform);

  // Load the source pref file
  nsCOMPtr<nsIPrefService> psvc(do_GetService(NS_PREFSERVICE_CONTRACTID));
  psvc->ResetPrefs();

  nsCOMPtr<nsIFile> sourcePrefsFile;
  mSourceProfile->Clone(getter_AddRefs(sourcePrefsFile));
  sourcePrefsFile->Append(aSourcePrefFileName);
  psvc->ReadUserPrefs(sourcePrefsFile);

  nsCOMPtr<nsIPrefBranch> branch(do_QueryInterface(psvc));
  for (transform = gTransforms; transform < end; ++transform)
    transform->prefGetterFunc(transform, branch);

  nsVoidArray* fontPrefs = new nsVoidArray();
  if (!fontPrefs)
    return NS_ERROR_OUT_OF_MEMORY;
  ReadFontsBranch(psvc, fontPrefs);

  // Now that we have all the pref data in memory, load the target pref file,
  // and write it back out
  psvc->ResetPrefs();
  for (transform = gTransforms; transform < end; ++transform)
    transform->prefSetterFunc(transform, branch);

  WriteFontsBranch(psvc, fontPrefs);
  delete fontPrefs;
  fontPrefs = nsnull;

  nsCOMPtr<nsIFile> targetPrefsFile;
  mTargetProfile->Clone(getter_AddRefs(targetPrefsFile));
  targetPrefsFile->Append(aTargetPrefFileName);
  psvc->SavePrefFile(targetPrefsFile);

  psvc->ResetPrefs();
  psvc->ReadUserPrefs(nsnull);

  return NS_OK;
}

struct FontPref {
  char*         prefName;
  PRInt32       type;
  union {
    char*       stringValue;
    PRInt32     intValue;
    PRBool      boolValue;
    PRUnichar*  wstringValue;
  };
};

void
nsFirefoxProfileMigrator::ReadFontsBranch(nsIPrefService* aPrefService, 
                                            nsVoidArray* aPrefs)
{
  // Enumerate the branch
  nsCOMPtr<nsIPrefBranch> branch;
  aPrefService->GetBranch("font.", getter_AddRefs(branch));

  PRUint32 count;
  char** prefs = nsnull;
  nsresult rv = branch->GetChildList("", &count, &prefs);
  if (NS_FAILED(rv)) return;

  for (PRUint32 i = 0; i < count; ++i) {
    // Save each pref's value into an array
    char* currPref = prefs[i];
    PRInt32 type;
    branch->GetPrefType(currPref, &type);
    FontPref* pref = new FontPref;
    pref->prefName = currPref;
    pref->type = type;
    switch (type) {
    case nsIPrefBranch::PREF_STRING:
      rv = branch->GetCharPref(currPref, &pref->stringValue);
      break;
    case nsIPrefBranch::PREF_BOOL:
      rv = branch->GetBoolPref(currPref, &pref->boolValue);
      break;
    case nsIPrefBranch::PREF_INT:
      rv = branch->GetIntPref(currPref, &pref->intValue);
      break;
    case nsIPrefBranch::PREF_INVALID:
      {
        nsCOMPtr<nsIPrefLocalizedString> str;
        rv = branch->GetComplexValue(currPref, 
                                    NS_GET_IID(nsIPrefLocalizedString), 
                                    getter_AddRefs(str));
        if (NS_SUCCEEDED(rv) && str)
          str->ToString(&pref->wstringValue);
      }
      break;
    }

    if (NS_SUCCEEDED(rv))
      aPrefs->AppendElement((void*)pref);
  }
}

void
nsFirefoxProfileMigrator::WriteFontsBranch(nsIPrefService* aPrefService,
                                             nsVoidArray* aPrefs)
{
  nsresult rv;

  // Enumerate the branch
  nsCOMPtr<nsIPrefBranch> branch;
  aPrefService->GetBranch("font.", getter_AddRefs(branch));

  PRUint32 count = aPrefs->Count();
  for (PRUint32 i = 0; i < count; ++i) {
    FontPref* pref = (FontPref*)aPrefs->ElementAt(i);
    switch (pref->type) {
    case nsIPrefBranch::PREF_STRING:
      rv = branch->SetCharPref(pref->prefName, pref->stringValue);
      PL_strfree(pref->stringValue);
      pref->stringValue = nsnull;
      break;
    case nsIPrefBranch::PREF_BOOL:
      rv = branch->SetBoolPref(pref->prefName, pref->boolValue);
      break;
    case nsIPrefBranch::PREF_INT:
      rv = branch->SetIntPref(pref->prefName, pref->intValue);
      break;
    case nsIPrefBranch::PREF_INVALID:
      nsCOMPtr<nsIPrefLocalizedString> pls(do_CreateInstance("@mozilla.org/pref-localizedstring;1"));
      pls->SetData(pref->wstringValue);
      rv = branch->SetComplexValue(pref->prefName, 
                                   NS_GET_IID(nsIPrefLocalizedString),
                                   pls);
      NS_Free(pref->wstringValue);
      pref->wstringValue = nsnull;
      break;
    }
    NS_Free(pref->prefName);
    pref->prefName = nsnull;
    delete pref;
    pref = nsnull;
  }
  aPrefs->Clear();
}

nsresult
nsFirefoxProfileMigrator::CopyPreferences(PRBool aReplace)
{
  printf("****** nsFirefoxProfileMigrator::CopyPreferences() - Migrating Settings\n");
  nsresult rv = NS_OK;
  if (!aReplace)
    return rv;

  // Prefs files
  rv |= TransformPreferences(FILE_NAME_PREFS, FILE_NAME_PREFS);
  rv |= CopyFile(FILE_NAME_USER_PREFS, FILE_NAME_USER_PREFS);

  // Security Stuff
  rv |= CopyFile(FILE_NAME_CERT8DB, FILE_NAME_CERT8DB);
  rv |= CopyFile(FILE_NAME_SECMODDB, FILE_NAME_SECMODDB);

  nsAutoString fileName;
  rv |= GetAppFileName(fileName);
  if (!fileName.EqualsLiteral(NETSCAPE8_FILENAME)) {
    // for Netscape 8, the file containing the master password crashes
    // the application when importing passwords. We don't need anyway.
    rv |= CopyFile(FILE_NAME_KEY3DB, FILE_NAME_KEY3DB);
  }

  // User MIME Type overrides
  rv |= CopyFile(FILE_NAME_MIMETYPES, FILE_NAME_MIMETYPES);

  rv |= CopyUserStyleSheets();
  return rv;
}

nsresult 
nsFirefoxProfileMigrator::CopyUserStyleSheets()
{
  nsresult rv = NS_OK;

  nsCOMPtr<nsIFile> sourceUserContent;
  mSourceProfile->Clone(getter_AddRefs(sourceUserContent));
  sourceUserContent->Append(DIR_NAME_CHROME);
  sourceUserContent->Append(FILE_NAME_USERCONTENT);

  PRBool exists = PR_FALSE;
  sourceUserContent->Exists(&exists);
  if (exists) {
    nsCOMPtr<nsIFile> targetUserContent;
    mTargetProfile->Clone(getter_AddRefs(targetUserContent));
    targetUserContent->Append(DIR_NAME_CHROME);
    nsCOMPtr<nsIFile> targetChromeDir;
    targetUserContent->Clone(getter_AddRefs(targetChromeDir));
    targetUserContent->Append(FILE_NAME_USERCONTENT);

    targetUserContent->Exists(&exists);
    if (exists)
      targetUserContent->Remove(PR_FALSE);

    rv |= sourceUserContent->CopyTo(targetChromeDir, FILE_NAME_USERCONTENT);
  }

  nsCOMPtr<nsIFile> sourceUserChrome;
  mSourceProfile->Clone(getter_AddRefs(sourceUserChrome));
  sourceUserChrome->Append(DIR_NAME_CHROME);
  sourceUserChrome->Append(FILE_NAME_USERCHROME);

  sourceUserChrome->Exists(&exists);
  if (exists) {
    nsCOMPtr<nsIFile> targetUserChrome;
    mTargetProfile->Clone(getter_AddRefs(targetUserChrome));
    targetUserChrome->Append(DIR_NAME_CHROME);
    nsCOMPtr<nsIFile> targetChromeDir;
    targetUserChrome->Clone(getter_AddRefs(targetChromeDir));
    targetUserChrome->Append(FILE_NAME_USERCHROME);

    targetUserChrome->Exists(&exists);
    if (exists)
      targetUserChrome->Remove(PR_FALSE);

    rv |= sourceUserChrome->CopyTo(targetChromeDir, FILE_NAME_USERCHROME);
  }

  return rv;
}

nsresult
nsFirefoxProfileMigrator::CopyCookies(PRBool aReplace)
{
  printf("****** nsFirefoxProfileMigrator::CopyCookies() - Migrating Cookies\n");
  nsresult rv;
  if (aReplace)
    rv = CopyFile(FILE_NAME_COOKIES, FILE_NAME_COOKIES);
  else {
    nsCOMPtr<nsIFile> seamonkeyCookiesFile;
    mSourceProfile->Clone(getter_AddRefs(seamonkeyCookiesFile));
    seamonkeyCookiesFile->Append(FILE_NAME_COOKIES);

    rv = ImportNetscapeCookies(seamonkeyCookiesFile);
  }
  return rv;
}

nsresult
nsFirefoxProfileMigrator::CopyHistory(PRBool aReplace)
{
  printf("****** nsFirefoxProfileMigrator::CopyHistory() - Migrating Browsing History\n");
  nsresult rv = NS_OK;
  
  if (aReplace) {
    // First import so just copy the history since we have none yet
    rv = CopyFile(FILE_NAME_HISTORY, FILE_NAME_HISTORY);
  } else {
    // TODO: Read and append browsing history.
  }
  
  return rv;
}

nsresult
nsFirefoxProfileMigrator::CopyFormHistory(PRBool aReplace)
{
  printf("****** nsFirefoxProfileMigrator::CopyFormHistory() - Migrating Form History\n");
  nsresult rv = NS_OK;
  if (aReplace) {
    // First import so just copy the history since we have none yet
    rv = CopyFile(FILE_NAME_FORMHISTORY, FILE_NAME_FORMHISTORY);
  } else {
    // Check for existence of "from" form history file.
    nsCOMPtr<nsIFile> historyFile;
    mSourceProfile->Clone(getter_AddRefs(historyFile));
    historyFile->Append(FILE_NAME_FORMHISTORY);
    PRBool exists = PR_FALSE;
    historyFile->Exists(&exists);
    if (exists) {
      // Get the "to" form history service for additions.
      nsCOMPtr<nsIFormHistory2> formHistory(do_GetService("@mozilla.org/satchel/form-history;1"));
      if (formHistory) {
        // Open the "from" Mork DB.
        nsCAutoString filePath;
        historyFile->GetNativePath(filePath);
        if (NS_SUCCEEDED(OpenFormHistoryDB(filePath.get()))) {
          nsCOMPtr<nsIMdbTableRowCursor> rowCursor;
          mdb_err err = mTable->GetTableRowCursor(mEnv, -1, getter_AddRefs(rowCursor));
          NS_ENSURE_TRUE(!err, NS_ERROR_FAILURE);
          nsCOMPtr<nsIMdbRow> row;
          mdb_pos pos;
          nsAutoString name;
          nsAutoString value;
          do {
            rowCursor->NextRow(mEnv, getter_AddRefs(row), &pos);
            if (!row)
              break;

            GetRowValue(row, kToken_NameColumn, name);
            GetRowValue(row, kToken_ValueColumn, value);
            formHistory->AddEntry(name, value);

          } while (1);

          nsCOMPtr<nsIMdbThumb> thumb;
          err = mStore->CompressCommit(mEnv, getter_AddRefs(thumb));
          if (!err)
            err = UseThumb(thumb, nsnull);
        }
        // Do this regardless to clean up on failure of open.
        CloseFormHistoryDB();
      }
    }
  }

  return rv;
}

nsresult
nsFirefoxProfileMigrator::OpenFormHistoryDB(const char * aPath)
{
  nsresult rv = NS_OK;
  if (mStore)
    return rv;

  // Get an Mdb Factory
  static NS_DEFINE_CID(kMorkCID, NS_MORK_CID);
  nsCOMPtr<nsIMdbFactoryFactory> mdbFactory = do_CreateInstance(kMorkCID, &rv);
  NS_ENSURE_SUCCESS(rv, rv);
  rv = mdbFactory->GetMdbFactory(getter_AddRefs(mMdbFactory));
  NS_ENSURE_SUCCESS(rv, rv);

  // Create the Mdb environment
  mdb_err err = mMdbFactory->MakeEnv(nsnull, &mEnv);
  NS_ASSERTION(err == 0, "ERROR: Unable to create Form History mdb");
  mEnv->SetAutoClear(PR_TRUE);
  NS_ENSURE_TRUE(!err, NS_ERROR_FAILURE);
  mEnv->SetErrorHook(new MigrationErrorHook());

  nsCOMPtr<nsIMdbFile> oldFile;
  nsIMdbHeap* dbHeap = 0;
  err = mMdbFactory->OpenOldFile(mEnv, dbHeap, aPath, mdbBool_kFalse, getter_AddRefs(oldFile));
  NS_ENSURE_TRUE(!err && oldFile, NS_ERROR_FAILURE);

  mdb_bool canOpen = 0;
  mdbYarn outFormat = {nsnull, 0, 0, 0, 0, nsnull};
  err = mMdbFactory->CanOpenFilePort(mEnv, oldFile, &canOpen, &outFormat);
  NS_ENSURE_TRUE(!err && canOpen, NS_ERROR_FAILURE);

  nsCOMPtr<nsIMdbThumb> thumb;
  mdbOpenPolicy policy = {{0, 0}, 0, 0};
  err = mMdbFactory->OpenFileStore(mEnv, dbHeap, oldFile, &policy, getter_AddRefs(thumb));
  NS_ENSURE_TRUE(!err && thumb, NS_ERROR_FAILURE);

  PRBool done;
  mdb_err thumbErr = UseThumb(thumb, &done);

  if (err == 0 && done)
    err = mMdbFactory->ThumbToOpenStore(mEnv, thumb, &mStore);
  NS_ENSURE_TRUE(!err, NS_ERROR_FAILURE);

  err = mStore->StringToToken(mEnv, "ns:formhistory:db:row:scope:formhistory:all", &kToken_RowScope);
  if (err != 0) return NS_ERROR_FAILURE;

  err = mStore->StringToToken(mEnv, "Value", &kToken_ValueColumn);
  if (err != 0) return NS_ERROR_FAILURE;

  err = mStore->StringToToken(mEnv, "Name", &kToken_NameColumn);
  if (err != 0) return NS_ERROR_FAILURE;

  err = mStore->StringToToken(mEnv, "ByteOrder", &kToken_ByteOrder);
  if (err != 0) return NS_ERROR_FAILURE;

  mdbOid oid = {kToken_RowScope, 1};
  err = mStore->GetTable(mEnv, &oid, &mTable);
  NS_ENSURE_TRUE(!err, NS_ERROR_FAILURE);
  if (!mTable) {
    NS_WARNING("ERROR: Form history file is corrupt, now deleting it.");
    return NS_ERROR_FAILURE;
  }

  err = mTable->GetMetaRow(mEnv, &oid, nsnull, getter_AddRefs(mMetaRow));
  if (err)
    NS_WARNING("Could not get meta row");

  if (NS_FAILED(thumbErr)) {
    rv = NS_ERROR_FAILURE;
  } else {
    rv = InitByteOrder();
  }

  return rv;
}

void
nsFirefoxProfileMigrator::CloseFormHistoryDB(void)
{
  mMetaRow = nsnull;

  if (mTable)
    mTable->Release();

  if (mStore)
    mStore->Release();

  if (mEnv)
    mEnv->Release();

  mTable = nsnull;
  mEnv = nsnull;
  mStore = nsnull;
}

nsresult
nsFirefoxProfileMigrator::GetRowValue(nsIMdbRow *aRow, mdb_column aCol, nsAString &aValue)
{
  mdbYarn yarn;
  mdb_err err = aRow->AliasCellYarn(mEnv, aCol, &yarn);
  if (err != 0)
    return NS_ERROR_FAILURE;

  aValue.Truncate(0);
  if (!yarn.mYarn_Fill)
    return NS_OK;
  
  switch (yarn.mYarn_Form) {
    case 0: { // unicode
      PRUint32 len = yarn.mYarn_Fill / sizeof(PRUnichar);
      if (mReverseByteOrder) {
        PRUnichar *swapval = new PRUnichar[len];
        if (!swapval)
          return NS_ERROR_OUT_OF_MEMORY;
        SwapBytes(swapval, (const PRUnichar*)yarn.mYarn_Buf, len);
        aValue.Assign(swapval, len);
        delete swapval;
      }
      else
        aValue.Assign((const PRUnichar *)yarn.mYarn_Buf, len);
      break;
    }
    default:
      return NS_ERROR_UNEXPECTED;
  }
  
  return NS_OK;
}

nsresult
nsFirefoxProfileMigrator::UseThumb(nsIMdbThumb *aThumb, PRBool *aDone)
{
  mdb_count total;
  mdb_count current;
  mdb_bool done;
  mdb_bool broken;
  mdb_err err;
  
  do {
    err = aThumb->DoMore(mEnv, &total, &current, &done, &broken);
  } while ((err == 0) && !broken && !done);
  
  if (aDone)
    *aDone = done;
  
  return err ? NS_ERROR_FAILURE : NS_OK;
}

nsresult
nsFirefoxProfileMigrator::InitByteOrder(void)
{
  // bigEndian and littleEndian are endianness markers that are stored in
  // the formhistory db as UTF-16.  Define them to be strings easily
  // recognized in either endianness.
  nsAutoString bigEndianByteOrder((PRUnichar*)"BBBB", 2);
  nsAutoString littleEndianByteOrder((PRUnichar*)"llll", 2);
#ifdef IS_BIG_ENDIAN
  nsAutoString nativeByteOrder(bigEndianByteOrder);
#else
  nsAutoString nativeByteOrder(littleEndianByteOrder);
#endif

  nsresult rv = NS_OK;
  nsAutoString fileByteOrder;
  mdb_err err = GetRowValue(mMetaRow, kToken_ByteOrder, fileByteOrder);
  if (err)
    rv = NS_ERROR_FAILURE;
  else
    mReverseByteOrder = !fileByteOrder.Equals(nativeByteOrder);

  return rv;
}

nsresult
nsFirefoxProfileMigrator::CopyPasswords(PRBool aReplace)
{
  printf("****** nsFirefoxProfileMigrator::CopyPasswords() - Migrating Passwords\n");
  nsresult rv;

  nsCOMPtr<nsIPrefService> psvc(do_GetService(NS_PREFSERVICE_CONTRACTID));

  if (aReplace) {
    // Ensure in-memory prefs are from target profile, not source profile.
    psvc->ResetPrefs();
    nsCOMPtr<nsIFile> targetPrefFile;
    mTargetProfile->Clone(getter_AddRefs(targetPrefFile));
    targetPrefFile->Append(FILE_NAME_PREFS);
    psvc->ReadUserPrefs(targetPrefFile);
  }
  nsCOMPtr<nsIPrefBranch> branch(do_QueryInterface(psvc));

  // Find out what the signons file is called in Flock.
  nsXPIDLCString signonsFileName2;
  rv = branch->GetCharPref("signon.SignonFileName2", getter_Copies(signonsFileName2));

  if (signonsFileName2.IsEmpty())
    return NS_OK;

  nsAutoString fileName2;
  fileName2.AssignWithConversion(signonsFileName2);

  if (SourceFileExists(mSourceSignonFileName)) {
    if (aReplace) {
      // Overwrite.
      rv = CopyFile(mSourceSignonFileName, fileName2);
    } else {
      // Merge.
      nsCOMPtr<nsIFile> firefoxPasswordsFile;
      mSourceProfile->Clone(getter_AddRefs(firefoxPasswordsFile));
      firefoxPasswordsFile->Append(mSourceSignonFileName);

      nsCOMPtr<nsIPasswordManagerInternal> pmi(do_GetService("@mozilla.org/passwordmanager;1"));
      rv = pmi->ReadPasswords(firefoxPasswordsFile);  // Firefox passwords to merge from
      if (NS_SUCCEEDED(rv)) {
        nsCOMPtr<nsIPasswordManager> pm = do_QueryInterface(pmi);
        if (pm) {
          pm->AddUser(   NS_LITERAL_CSTRING("http://nosuchhost.flock.com/"), NS_LITERAL_STRING("nosuchuser"), NS_LITERAL_STRING("password"));
          pm->RemoveUser(NS_LITERAL_CSTRING("http://nosuchhost.flock.com/"), NS_LITERAL_STRING("nosuchuser"));
        }
      }
    }
  }
  return rv;
}

nsresult
nsFirefoxProfileMigrator::CopyBookmarks(PRBool aReplace)
{
  printf("****** nsFirefoxProfileMigrator::CopyBookmarks() - Migrating Bookmarks->Favorites\n");
  if (aReplace)
      return CopyFile(FILE_NAME_BOOKMARKS, FILE_NAME_BOOKMARKS);

  nsAutoString sourceNameKey;
  nsresult rv = GetBookmarksSourceName(sourceNameKey);
  NS_ENSURE_SUCCESS(rv, rv);

  return ImportNetscapeBookmarks(FILE_NAME_BOOKMARKS, sourceNameKey.get());
}

nsresult
nsFirefoxProfileMigrator::CopyOtherData(PRBool aReplace)
{
  printf("****** nsFirefoxProfileMigrator::CopyOtherData() - Migrating OtherData\n");

  if (!aReplace)
    return NS_OK;

  nsresult rv = NS_OK;
  rv |= CopyFile(FILE_NAME_DOWNLOADS, FILE_NAME_DOWNLOADS);
//  rv |= CopyFile(FILE_NAME_SEARCH, FILE_NAME_SEARCH); bad on osx
//  rv |= CopyFile(FILE_NAME_LOCALSTORE, FILE_NAME_LOCALSTORE);// avoid the buttonset problem ;)

  return rv;
}

void
nsFirefoxProfileMigrator::MigrateYahooToolbar(void) {
  // First Check if Yahoo Toolbar is installed.
  nsresult rv;
  PRBool doInstall = PR_FALSE;
  
  nsCOMPtr<nsIFile> extensionFolder;
  mSourceProfile->Clone(getter_AddRefs(extensionFolder));
  extensionFolder->Append(DIR_NAME_EXTENSIONS);
  
  // Now we have the directory of the extensions, we need to go through them until we find the Yahoo Toolbar or have none left
  nsCOMPtr<nsISimpleEnumerator> entries;
  rv = extensionFolder->GetDirectoryEntries(getter_AddRefs(entries));
  if (NS_FAILED(rv)) return;
  
  do { // loop through the funny named directories
    PRBool hasMore = PR_FALSE;
    rv = entries->HasMoreElements(&hasMore);
    if (NS_FAILED(rv) || !hasMore) break;

    nsCOMPtr<nsISupports> supp;
    rv = entries->GetNext(getter_AddRefs(supp));
    if (NS_FAILED(rv)) break;

    nsCOMPtr<nsIFile> currFile(do_QueryInterface(supp));
	  if (!currFile) continue;
    currFile->AppendNative(NS_LITERAL_CSTRING("install.rdf"));        
    // Check the install.rdf and look for a <em:name>Yahoo! Toolbar</em:name> match
    // since this is a rdf we can load it in an rdfDatasource :)
    nsCAutoString urlSpec;
    rv = NS_GetURLSpecFromFile(currFile, urlSpec);
    if (NS_FAILED(rv)) continue;

    // Load the datasource from install.rdf
    nsCOMPtr<nsIRDFDataSource> ds;
    nsCOMPtr<nsIRDFService> rdfs(do_GetService("@mozilla.org/rdf/rdf-service;1"));
	  if (!rdfs) continue;

    rv = rdfs->GetDataSourceBlocking(urlSpec.get(), getter_AddRefs(ds));
    if (NS_FAILED(rv)) continue;
    
    nsCOMPtr<nsIRDFResource> installManifest;
    nsCOMPtr<nsIRDFResource> emName;
    nsCOMPtr<nsIRDFNode> emNameNode;
    rv = rdfs->GetResource(INSTALL_MANIFEST, getter_AddRefs(installManifest));
    if (NS_FAILED(rv)) continue;
    rv = rdfs->GetResource(EMNAME_RESOURCE, getter_AddRefs(emName));
    if (NS_FAILED(rv)) continue;
    rv = ds->GetTarget(installManifest, emName, PR_TRUE, getter_AddRefs(emNameNode));
    if (NS_FAILED(rv)) continue;
    nsCOMPtr<nsIRDFLiteral> emNameLiteral(do_QueryInterface(emNameNode));
	  if (!emNameLiteral) continue;

    // We now have the value of <em:name in emNameValue...check that it matches
    const PRUnichar *emValue;
    rv = emNameLiteral->GetValueConst(&emValue);
	  if (NS_FAILED(rv) || !emValue) continue;
    
    nsCAutoString extName;
    CopyUTF16toUTF8(emValue, extName);
    if (extName.Equals("Yahoo! Toolbar")) {
      doInstall = true;
      break;
    }   
  } while(1);
  
  if (!doInstall) {
    return;
  }
  
  // Install the extension
  nsCOMPtr<nsIExtensionManager> em = do_GetService("@mozilla.org/extensions/manager;1", &rv);
  if (NS_FAILED(rv)) return;
  
  nsCOMPtr<nsILocalFile> xpiFile;
  if (NS_FAILED(rv)) return;

  // We need to get the current install location and use that with the yahootoolbar.xpi  
  nsCOMPtr<nsIProperties> dirSvc = do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID, &rv);
  if (!dirSvc) return;
  dirSvc->Get(NS_OS_CURRENT_PROCESS_DIR, NS_GET_IID(nsIFile), getter_AddRefs(xpiFile));
  xpiFile->AppendNative(NS_LITERAL_CSTRING("yahootoolbar.xpi"));
  
  PRBool xpiExists = PR_FALSE;
  xpiFile->Exists(&xpiExists);
  
  if (xpiExists) {
    em->InstallItemFromFile(xpiFile, NS_INSTALL_LOCATION_APPPROFILE);
  }

}

nsresult nsFirefoxProfileMigrator::GetBookmarksSourceName(nsAString &aName)
{
  aName.Assign(NS_LITERAL_STRING("sourceNameFirefox"));
  return NS_OK;
}

nsresult nsFirefoxProfileMigrator::GetAppFileName(nsAString &aFileName)
{
#ifdef XP_WIN
  aFileName.AssignLiteral("firefox.exe");
#elif defined(XP_MACOSX)
  aFileName.AssignLiteral("Firefox");
#elif defined(XP_UNIX)
  aFileName.AssignLiteral("firefox");
#endif
  return NS_OK;
}

nsresult nsNetscape9ProfileMigrator::AppendNativeProfilePath(nsILocalFile *aFile)
{
#ifdef XP_WIN
  aFile->Append(NS_LITERAL_STRING("Netscape"));
  aFile->Append(NS_LITERAL_STRING("Navigator"));
#elif defined(XP_MACOSX)
  aFile->Append(NS_LITERAL_STRING("Application Support"));
  aFile->Append(NS_LITERAL_STRING("Navigator"));
#elif defined(XP_UNIX)
  aFile->Append(NS_LITERAL_STRING(".netscape"));
  aFile->Append(NS_LITERAL_STRING("navigator"));
#endif
  return NS_OK;
}

nsresult nsNetscape9ProfileMigrator::GetBookmarksSourceName(nsAString &aName)
{
  aName.Assign(NS_LITERAL_STRING("sourceNameNetscape9"));
  return NS_OK;
}

nsresult nsNetscape9ProfileMigrator::GetAppFileName(nsAString &aFileName)
{
#ifdef XP_WIN
  aFileName.AssignLiteral("navigator.exe");
#elif defined(XP_MACOSX)
  aFileName.AssignLiteral("Navigator");
#elif defined(XP_UNIX)
  aFileName.AssignLiteral("navigator");
#endif
  return NS_OK;
}

#if defined(XP_WIN)

nsresult nsNetscape8ProfileMigrator::AppendNativeProfilePath(nsILocalFile *aFile)
{
  aFile->Append(NS_LITERAL_STRING("Netscape"));
  aFile->Append(NS_LITERAL_STRING("NSB"));
  return NS_OK;
}

nsresult nsNetscape8ProfileMigrator::GetBookmarksSourceName(nsAString &aName)
{
  aName.Assign(NS_LITERAL_STRING("sourceNameNetscape8"));
  return NS_OK;
}

nsresult nsNetscape8ProfileMigrator::GetAppFileName(nsAString &aFileName)
{
  aFileName.AssignLiteral(NETSCAPE8_FILENAME);
  return NS_OK;
}

nsresult nsNetscape8ProfileMigrator::GetSourceSignonFileName(PRBool aReplace,
                                                             char** aFileName)
{
  *aFileName = ToNewCString(NS_LITERAL_CSTRING("c0g1n3n5.r7u"));
  return NS_OK;
}

static const char kRealmSeparator[] = ".";
#define PASSCARD_VERSION_INDEX 17

nsresult nsNetscape8ProfileMigrator::CopyPasswords(PRBool aReplace)
{
  // copy from Netscape 8 Passcards
  nsCOMPtr<nsIFile> countersignFile;
  mSourceProfile->Clone(getter_AddRefs(countersignFile));
  countersignFile->Append(mSourceSignonFileName);

  nsCOMPtr<nsIInputStream> fileStream;
  NS_NewLocalFileInputStream(getter_AddRefs(fileStream), countersignFile);
  if (!fileStream)
    return NS_ERROR_OUT_OF_MEMORY;

  nsCOMPtr<nsILineInputStream> lineStream = do_QueryInterface(fileStream);
  NS_ASSERTION(lineStream, "File stream is not an nsILineInputStream");

  nsCAutoString realm, utf8Buffer; // MERC JVL

  // Go through preferences
  PRBool moreData = PR_FALSE;
  nsresult rv;
  int lineIndex = 0;
  enum { STATE_REJECT, STATE_REALM, STATE_USERFIELD, STATE_USERVALUE,
         STATE_PASSFIELD, STATE_PASSVALUE, STATE_PROTECT = 9 };
  PRInt32 state = STATE_REJECT;
  PRInt32 STATE_ENDPASSCARD = 12; // for Netscape 8.0
  do {
    rv = ReadObfuscatedLine(lineStream, utf8Buffer, &moreData);
    if (NS_FAILED(rv))
      return NS_OK;
    lineIndex++;
  } while (!utf8Buffer.Equals(NS_LITERAL_CSTRING("|")) && moreData);

  if (lineIndex == PASSCARD_VERSION_INDEX) {
    // this pref line is only found in Netscape 8.1
    STATE_ENDPASSCARD += 4; // 4 passcard fields added in Netscape 8.1
  }

  // Now read the passcards
  nsCOMPtr<nsIPasswordManagerInternal> pmi;
  pmi = do_GetService("@mozilla.org/passwordmanager;1");
  nsAutoString userField, userValue, passField, passValue;
  PRBool isProtected;
  do {
    rv = ReadObfuscatedLine(lineStream, utf8Buffer, &moreData);
    if (NS_FAILED(rv))
      return NS_OK;

    switch (state++) {
    case STATE_REJECT:
      if (utf8Buffer.Equals(NS_LITERAL_CSTRING(":"))) {
        break;
      }
      state = STATE_REJECT;
      // fall through STATE_REALM
    case STATE_REALM: {
        // Workaround for the pb of Netscape 8 stripping www
        PRInt32 posDot = utf8Buffer.FindChar('.');
        if(posDot == -1 || utf8Buffer.FindChar('.', posDot + 1) == -1) {
          // since there're less than 2 dots in the realm we have to guess
          // that Netscape 8 stripped www
          utf8Buffer = NS_LITERAL_CSTRING("www.") + utf8Buffer;
        }
        if (state == STATE_REJECT) {
          nsCOMPtr<nsIPasswordManager> pm = do_QueryInterface(pmi);
          pm->AddReject(NS_LITERAL_CSTRING("http://") + utf8Buffer);
        } else {
          realm.Assign(NS_LITERAL_CSTRING("http://") + utf8Buffer);
        }
      }
      break;
    case STATE_USERFIELD:
      // Add previous entry
      if (realm != NS_LITERAL_CSTRING("http://www.DefaultPasscard")) {
        // Weed out empty username+password entries
        if ((!userValue.IsEmpty() || !passValue.IsEmpty()) && !isProtected) {
          pmi->AddUserFull(realm, userValue, passValue, userField, passField);
        }
      }
      // If the line is a ., we've reached the end of this realm's entries.
      if (utf8Buffer.Equals(kRealmSeparator)) {
        state = STATE_REALM;
        userValue.AssignLiteral("");
        passValue.AssignLiteral("");
      } else {
        CopyUTF8toUTF16(utf8Buffer, userField);
      }
      break;
    case STATE_USERVALUE:
      CopyUTF8toUTF16(utf8Buffer, userValue);
      break;
    case STATE_PASSFIELD:
      // Strip off the leading "*" character
      CopyUTF8toUTF16(Substring(utf8Buffer, 1, utf8Buffer.Length() - 1),
                      passField);
      break;
    case STATE_PASSVALUE:
      CopyUTF8toUTF16(utf8Buffer, passValue);
      break;
    case STATE_PROTECT:
      isProtected = utf8Buffer.EqualsLiteral("1");
      break;
    default:
      if (state == STATE_ENDPASSCARD) {
        state = STATE_USERFIELD;
      }
    }
  } while (moreData);

  fileStream->Close();
  return NS_OK;
}

#define SC_TRUST_MAGIC_NUMBER	  14
// XOR the buffer to obfuscate/unobfuscate the information
nsresult nsNetscape8ProfileMigrator::ReadObfuscatedLine(nsILineInputStream* lis,
                                                        nsACString& buffer,
                                                        PRBool* moreData)
{
  nsresult rv = lis->ReadLine(buffer, moreData);
  if (NS_SUCCEEDED(rv)) {
    nsCString::iterator substr_start, substr_end;
    buffer.BeginWriting(substr_start);
    buffer.EndWriting(substr_end);
    while (substr_start != substr_end) {
      *substr_start = ( SC_TRUST_MAGIC_NUMBER ^ *substr_start );
      ++substr_start;
    }
  }
  return rv;
}

#endif
