/*
 * BEGIN FLOCK GPL
 *
 * Copyright Flock Inc. 2005-2008
 * http://flock.com
 *
 * This file may be used under the terms of the
 * GNU General Public License Version 2 or later (the "GPL"),
 * http://www.gnu.org/licenses/gpl.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * END FLOCK GPL
 */

#include "flockLuceneThread.h"

#include "flockLuceneCID.h"

#include "nsIEventQueueService.h"

#include "nsIComponentManager.h"
#include "nsIServiceManager.h"
#include "nsServiceManagerUtils.h"
#include "nsIProxyObjectManager.h"

#include "prlock.h"


static NS_DEFINE_CID(kFlockLuceneImplCID, FLOCK_LUCENE_IMPL_CID);

NS_IMPL_THREADSAFE_ISUPPORTS1(flockLuceneThread, nsIRunnable)

flockLuceneThread::flockLuceneThread()
  : nsIRunnable(), mIndexDir(NULL)
{
  /* member initializers and constructor code */
  mLock = PR_NewLock();
  mMonitor = PR_NewMonitor();
}

flockLuceneThread::~flockLuceneThread()
{
  /* destructor code */
  PR_DestroyLock(mLock);
  PR_DestroyMonitor(mMonitor);
  mIndexDir = NULL;
}

nsresult flockLuceneThread::Init(nsIFile *aIndexDir)
{
  if (!mThread) {
	  mIndexDir = aIndexDir;
	
	  // start the thread
	  nsresult rv = NS_NewThread(getter_AddRefs(mThread),
	                  NS_STATIC_CAST(nsIRunnable*, this),
	                  0,
	                  PR_UNJOINABLE_THREAD,
	                  PR_PRIORITY_LOW,
	                  PR_GLOBAL_THREAD);
	  NS_ASSERTION(NS_SUCCEEDED(rv), "failed to create thread for flockLuceneThread");
	  if (NS_FAILED(rv)) {
	    return rv;
	  }
	
	  // wait for the proxy to be created
	  PR_EnterMonitor(mMonitor);
	  if (!mProxySync) { // XXX race?
	    PR_Wait(mMonitor, PR_INTERVAL_NO_TIMEOUT);
	  }
	  PR_ExitMonitor(mMonitor);
  }

  return NS_OK;
}


nsresult flockLuceneThread::GetProxyAsync(flockILucene **aProxy)
{
  *aProxy = mProxyAsync;
  return NS_OK;
}

nsresult flockLuceneThread::GetProxySync(flockILucene **aProxy)
{
  *aProxy = mProxySync;
  return NS_OK;
}


NS_IMETHODIMP flockLuceneThread::Run()
{
  nsresult rv;

  nsCOMPtr<nsIEventQueue> currentThreadQ;

  PR_EnterMonitor(mMonitor);

  nsCOMPtr<nsIEventQueueService> eventService(do_GetService("@mozilla.org/event-queue-service;1", &rv));
  NS_ASSERTION(NS_SUCCEEDED(rv), "unable to get event queue service");
  if (NS_FAILED(rv)) {
    return rv;
  }

  rv = eventService->CreateMonitoredThreadEventQueue();
  NS_ASSERTION(NS_SUCCEEDED(rv), "unable to create event queue");
  if (NS_FAILED(rv)) {
    return rv;
  }

  rv = eventService->GetThreadEventQueue(NS_CURRENT_THREAD, getter_AddRefs(mEventQueue));
  NS_ASSERTION(NS_SUCCEEDED(rv), "unable to get current event queue");
  if (NS_FAILED(rv)) {
    return rv;
  }

  // create an impl object and get a proxy for it
  nsCOMPtr<flockILucene> luceneImpl = do_GetService(kFlockLuceneImplCID, &rv);
  if (NS_FAILED(rv)) {
    return rv;
  }
  
  rv = NS_GetProxyForObject(mEventQueue.get(),
                            NS_GET_IID(flockILucene),
                            luceneImpl,
                            PROXY_ASYNC | PROXY_ALWAYS,
                            getter_AddRefs(mProxyAsync));
  NS_ASSERTION(NS_SUCCEEDED(rv), "failed to create an async proxy for luceneImpl");
  if (NS_FAILED(rv)) {
    return rv;
  }
  
  rv = NS_GetProxyForObject(mEventQueue.get(),
                            NS_GET_IID(flockILucene),
                            luceneImpl,
                            PROXY_SYNC | PROXY_ALWAYS,
                            getter_AddRefs(mProxySync));
  NS_ASSERTION(NS_SUCCEEDED(rv), "failed to create an sync proxy for luceneImpl");
  if (NS_FAILED(rv)) {
    return rv;
  }

  // call init on the impl object
  mProxyAsync->Init(mIndexDir);

  // allow flockLuceneThread::Init to continue
  PR_Notify(mMonitor);
  PR_ExitMonitor(mMonitor);

  mRunning = true;
  PLEvent *event;
  while (mRunning) {
    rv = mEventQueue->WaitForEvent(&event);
    if (NS_FAILED(rv)) return rv;
    rv = mEventQueue->HandleEvent(event);
    if (NS_FAILED(rv)) return rv;
  }

  rv = eventService->DestroyThreadEventQueue();
  if (NS_FAILED(rv)) return rv;

  mEventQueue = nsnull;

  return NS_OK;
}


nsresult
flockLuceneThread::Shutdown()
{
// FIXME: lock?
  mRunning = false;
  return NS_OK;
}

