/* vim: ts=4 sw=4 expandtab cindent
*/
/*
 * BEGIN FLOCK GPL
 *
 * Copyright Flock Inc. 2005-2008
 * http://flock.com
 *
 * This file may be used under the terms of the
 * GNU General Public License Version 2 or later (the "GPL"),
 * http://www.gnu.org/licenses/gpl.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * END FLOCK GPL
 */

#include "flockAsyncImageScalerThread.h"

#include "flockAsyncImageScalerCID.h"

#include "nsIEventQueueService.h"

#include "nsIComponentManager.h"
#include "nsIServiceManager.h"
#include "nsServiceManagerUtils.h"
#include "nsIProxyObjectManager.h"

#include "prlock.h"
#include "nsMemory.h"

#include "nsAutoLock.h"

#include "nsCRT.h"

#include "stdio.h"

static NS_DEFINE_CID(kProxyObjectManagerCID, NS_PROXYEVENT_MANAGER_CID);
static NS_DEFINE_CID(kFlockAsyncImageScalerImplCID, FLOCK_ASYNCIMAGESCALER_IMPL_CID);

NS_IMPL_THREADSAFE_ISUPPORTS1(flockAsyncImageScalerThread, nsIRunnable)

flockAsyncImageScalerThread::flockAsyncImageScalerThread()
	: nsIRunnable()
{
    /* member initializers and constructor code */
    mLock = PR_NewLock();
    mMonitor = PR_NewMonitor();
}

flockAsyncImageScalerThread::~flockAsyncImageScalerThread()
{
    /* destructor code */
    PR_DestroyLock(mLock);
    PR_DestroyMonitor(mMonitor);

}

nsresult flockAsyncImageScalerThread::Init()
{
    if (!mThread) {

		// start the thread
        nsresult rv = NS_NewThread (getter_AddRefs (mThread),
									NS_STATIC_CAST(nsIRunnable*, this),
									0,
									PR_UNJOINABLE_THREAD,
									PR_PRIORITY_LOW,
									PR_GLOBAL_THREAD);
		NS_ASSERTION (NS_SUCCEEDED (rv), "failed to create thread for flockAsyncImageScalerThread");
        if (NS_FAILED (rv)) {
            return rv;
		}

		// wait for the proxy to be created
		PR_EnterMonitor (mMonitor);
		if (!mProxy) { // XXX race?
			PR_Wait (mMonitor, PR_INTERVAL_NO_TIMEOUT);
		}
		PR_ExitMonitor (mMonitor);
    }

    return NS_OK;
}


nsresult flockAsyncImageScalerThread::GetProxy(flockIAsyncImageScaler** aProxy)
{
	*aProxy = mProxy;
    return NS_OK;
}


NS_IMETHODIMP flockAsyncImageScalerThread::Run()
{
    nsresult rv;

    nsCOMPtr<nsIEventQueue> currentThreadQ;

    PR_EnterMonitor(mMonitor);

    nsCOMPtr<nsIEventQueueService> eventService (do_GetService ("@mozilla.org/event-queue-service;1", &rv));
    NS_ASSERTION (NS_SUCCEEDED (rv), "unable to get event queue service");
    if (NS_FAILED (rv)) {
		return rv;
	}

    rv = eventService->CreateMonitoredThreadEventQueue ();
    NS_ASSERTION (NS_SUCCEEDED (rv), "unable to create event queue");
    if (NS_FAILED (rv)) {
		return rv;
	}

    rv = eventService->GetThreadEventQueue (NS_CURRENT_THREAD, getter_AddRefs (mEventQueue));
    NS_ASSERTION (NS_SUCCEEDED (rv), "unable to get current event queue");
    if (NS_FAILED (rv)) {
		return rv;
	}

	// get the proxy manager
	nsCOMPtr<nsIProxyObjectManager> pmgr =
		do_GetService (kProxyObjectManagerCID, &rv);
    NS_ASSERTION (NS_SUCCEEDED (rv), "unable to get proxy manager service");
	if (NS_FAILED (rv)) {
		return rv;
	}

	// create an impl object and get a proxy for it
	flockIAsyncImageScaler* proxy = 0;
	rv = pmgr->GetProxy(mEventQueue.get(),
						 kFlockAsyncImageScalerImplCID,
						 (nsISupports*)nsnull, //nsISupports *aOuter, 
						 NS_GET_IID (flockIAsyncImageScaler),
						 PROXY_ASYNC|PROXY_ALWAYS,
						 //PROXY_SYNC|PROXY_ALWAYS,
						 (void**)&proxy);
	NS_ASSERTION (NS_SUCCEEDED (rv), 
			"failed to create a flockAsyncImageScalerImpl object and get a proxy for it");
	if (NS_FAILED (rv)) {
		return rv;
	}
	mProxy = proxy;

	// call init on the impl object
	//mProxy->Init();

	// allow flockAsyncImageScalerThread::Init to continue
    PR_Notify(mMonitor);
    PR_ExitMonitor(mMonitor);

	mRunning = true;
    PLEvent *event;
    while (mRunning) {
        rv = mEventQueue->WaitForEvent (&event);
        if (NS_FAILED (rv)) return rv;

        rv = mEventQueue->HandleEvent (event);
        if (NS_FAILED (rv)) return rv;
    }

    rv = eventService->DestroyThreadEventQueue ();
    if (NS_FAILED (rv)) return rv;

    mEventQueue = nsnull;

    return NS_OK;
}
